/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: FLUORfsub.c,v 1.5 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/

/* $Id: FLUORfsub.c,v 1.5 2002/05/08 21:34:35 jdejong Exp $ */
/*----------------------------------------------------------------------- */
/* Routines and parameters for FLUOR fringe fitting                       */
/*----------------------------------------------------------------------- */
#include "FLUORfsub.h"
#include "interpol.h"

/* global variables, possibly read from command line */
int      xxxphoto = 1;        /* make photometric corrections */

/* number of rows in input and last read */
long input_nrows = 0;
long input_last = 0;

/* photometric correction parameters */
int got_photom = 0;
float kappa11 = 1.0;
float kappa12 = 1.0;
float kappa21 = 1.0;
float kappa22 = 1.0;

/* global averaging of 1st moment */
float summom0 = 0.0;
long  cntmom0 = 0;

/* image_info for delay-rate image output */
image_info *FLUORimage = NULL;
float      *FLUORplane = NULL;  /* work array for image plane */
long        FOlast=0;        /* last word in FLUORplane written */
long        FOmax=0;         /* highest word in FLUORplane */
long        FOnx=0;          /* length of row in FLUORplane */

/* work arrays */
float FLwork[2048];
float FLspectrum[2048]; /* accumulated spectrum */
long  FLcount;

 
#define CLIGHT 2.997924562e8 /* speed of light */
#define TWOPI 6.283185308    /* two pi */

/* if DOSPLIT is defined split each scan into two parts */
/*#define DOSPLIT*/

image_info* init_FLUORpix(char *pixfile, fitsfile *inputfptr,
		       int nd, int nr, 
		       float del_d, float del_t, char *obs_date, int *ierr);
void FLUOR_moment (float mom[3], float *array, float center, int n, int wid);


/*----------------------------------------------------------------------- */
/******************* external functions **********************************/
/*----------------------------------------------------------------------- */

void setup_FLUORfringe_fit(char *pixfile, fitsfile *inputfptr, 
			   int nt, int nscan, int dophoto,
			   float OPD_incr,
			   imaging_detector_table_def* detector,
			   long input_n, long input_l)
/*----------------------------------------------------------------------- */
/*   Setup for fit MIDI Fourier Mode (undispersed) data                   */
/*   Necessary structures are allocated.                                  */
/*   Inputs:                                                              */
/*      pixfile   name of output Delay-rate image, NULL=>don't write      */
/*      inputfptr input file cfitsio pointer                              */
/*      nt        Number of samples per scan                              */
/*      nscan     Max. number of scans per solution                       */
/*      dophoto   if true (1) then apply photometric corrections          */
/*      time_incr Time step between samples in sec                        */
/*      ODP_incr  OPD step between samples in sec                         */
/*      detector  imaging_detector table definition structure             */
/*      input_nrows total number of rows in input                         */
/*      input_last  last row read in input                                */
/*----------------------------------------------------------------------- */
{
  int    mf, mt, jerr=0, i;
  float  del_d, del_r, del_t, ref_pix, df, deltar, deltad;
  double freq, freq0, freq_incr, freq1, freq2, lo_freq, delay, rate;
  
  /* save table row information */
  input_nrows = input_n;
  input_last = input_l;
  xxxphoto = dophoto;
  
  del_d = OPD_incr;
  del_t = 1.0;

  /* size of row depends on DOSPLIT */
  mt = nt;
#ifdef DOSPLIT
    mt = nt / 2;
#endif
  
  if (pixfile!=NULL) { /* Image requested? */
    FLUORimage = init_FLUORpix (pixfile, inputfptr, mt, nscan, del_d, del_t, 
			  detector->date_obs, &jerr);
    /* if there's a problem - turn off writing */
    if ((jerr!=0) || (FLUORimage==NULL)) {
      FLUORimage = NULL;
    }
  } /* end set up output images */

  /* init average spectrum */
  for (i=0; i<nt; i++) FLspectrum[i] = 0.0;
  FLcount = 0;

} /* end setup_FLUORfringe_fit */

void shutdown_FLUORfringe_fit(int *ierr)
/*----------------------------------------------------------------------- */
/*   Shutdown fit to FLUOR                                                */
/*   All allocated structures are deleted.                                */ 
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  int i, jerr=0;
  float amp, ph, sumre;
  
  /* write image */
  if (FLUORimage) {
    image_write(FLUORimage, FLUORplane, &jerr); /* write it out */
    /* if there's a problem - turn off writing */
    if (jerr!=0) {
      FLUORimage = NULL;
    }
  }
  
  if (FLUORimage) { /* image output? */
    if (*ierr) return; /* existing error? */
    image_close(FLUORimage, ierr);
    if (FLUORplane) free(FLUORplane);
  }

  /* print average spectrum */
  sumre = 0.0;
  if (FLcount>0) {
    for (i=0; i<FOnx/2; i++) {
      FLspectrum[i*2]   /= FLcount;
      FLspectrum[i*2+1] /= FLcount;
      sumre += FLspectrum[i*2];
      amp = sqrt (FLspectrum[i*2]*FLspectrum[i*2] + 
		  FLspectrum[i*2+1]*FLspectrum[i*2+1]);
      ph = atan2 (FLspectrum[i*2+1], FLspectrum[i*2]);
      fprintf (stdout,"avg channel %d %7.2f %6.1f\n", 
	       i, 10.0*amp, 57.296*ph);
    }
      fprintf (stdout,"sum of real part %7.2f\n", 
	       sumre);
  }

  /* print average momemt 1 */
  fprintf (stdout,"average moment 1 %f\n",  summom0/cntmom0);

  return;
} /* end shutdown_FOfringe_fit */


void get_FLUOR_scan(fitsfile *fptr, imaging_data_table_def* tabledef, 
		    imaging_data_row *row, int maxstep, 
		    float *OPD_off, float *times, float *IntChan,
		    long *ntime, int *ierr)
/*----------------------------------------------------------------------- */
/*   Loop reading a scan worth of FLUOR data                              */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  imaging_data table definition structure                 */
/*      row       imaging_data table row structure (last read)            */
/*      maxstep   maximum number of times in scan                         */
/*      OPD_off   OPDs (sec)                                              */
/*      times     times of entries in IntChan                             */
/*   Output:                                                              */
/*      row       imaging_data table row structure (last read)            */
/*      ntime     the actual number of times in IntChan                   */
/*      ierr      Error code: 0 => ok, checked on input, -1 => done       */
/*----------------------------------------------------------------------- */
{
  long irow, first, last, integ=0;
  int want, i, j;
  float xfirst[4];
  float sum11, sum21, sum12, sum22, sump1, sump2;
  float p11, p12, p21, p22, datai1, datai2, temp;

  /* set photometric ans interferometric columns in input */
  int i1col=0, i2col=1, p1col=2, p2col=3;
  
  first = input_last+1;
  last = input_nrows;

  sum11 = sum12 = sum21 = sum22 = 0.0;
  sump1 = sump2 = 0.0;
  /* loop over rows */
  for (irow = first; irow<=last; irow++) {

    /* sums for photometric corrections */
    if (xxxphoto) {
      /* problem here */
      if (row->opt_train[1]<=0) { /* shutter closed tele. 1 */
	got_photom = 0;
	sump2 +=  row->data[p2col][0];
	sum12 += row->data[i1col][0];
	sum22 += row->data[i2col][0];
      } else if (row->opt_train[0]<=0) { /* shutter closed tele. 2 */
	sump1 +=  row->data[p1col][0];
	sum11 += row->data[i1col][0];
	sum21 += row->data[i2col][0];
      }
    } /* end get photometric corrections */
  
    /* is the current one wanted? */
    /* on source with both shutters open */
    want = ((row->targtype[i1col]==1) &&
	    (row->opt_train[0]>0) &&
	    (row->opt_train[1]>0));

    /* redo photometric constants if necessary */
    if ((xxxphoto) && (want) && (!got_photom )) {
      got_photom = 1;
      if (sump1<1.0) sump1 = 1.0;
      if (sump2<1.0) sump2 = 1.0;
      kappa11 = sum11 / sump1;
      kappa12 = sum12 / sump2;
      kappa21 = sum21 / sump1;
      kappa22 = sum22 / sump2;
    } /* end get kappa array */

    /* is this the end of the scan? */
    if (want && (row->step<integ)) break;
    /* try breaking scans in half */

    /* split into two? */
#ifdef DOSPLIT
    if (want && ((integ)>=maxstep/2)) break;
#endif
    
    if (want) { /* if wanted copy data to work arrays */

      /* make photometric corrections? */
      if (xxxphoto) {
	p11 =  kappa11 * row->data[p1col][0];
	p21 =  kappa21 * row->data[p1col][0];
	p12 =  kappa12 * row->data[p2col][0];
	p22 =  kappa22 * row->data[p2col][0];
	if (p11<=0.0) p11 = 1.0;
	if (p12<=0.0) p12 = 1.0;
	if (p21<=0.0) p21 = 1.0;
	if (p22<=0.0) p22 = 1.0;
	datai1 = (row->data[i1col][0] - p11 - p12) /
	  sqrt (p11 * p12);
	datai2 = (row->data[i2col][0] - p21 - p22) /
	  sqrt (p21 * p22);
      } /* end apply photometric corrections */
      else { /* no photometric correction */
	datai1 = row->data[i1col][0];
	datai2 = row->data[i2col][0];
      }

      /* debug */
      if (row->targtype[i1col]==1) {
	if (row->step==1) {
	  xfirst[0] = datai1;
	  xfirst[1] = datai2;
	  xfirst[2] = row->data[p1col][0];
	  xfirst[3] = row->data[p2col][0];
	}
	
      /* debug */
	/*fprintf (stdout,"%3.3d %9.1f %9.1f %9.1f  %5d %5d  %5d\n", 
		 row->step, datai1*1000.0, datai2*1000.0,
		 datai1*1000.0-datai2*1000.0,
		 row->data[p1col][0],
		 row->data[p2col][0],
		 want);*/
	/*fprintf (stdout," %3.3d.00000 %9.1f    1    1 \n", 
	  row->step, datai1*1000-datai2*1000.0);*/
      }
      
      OPD_off[integ] = row->local_opd[0]; /* modulation delay */
      times[integ] = row->time;        /* time */
      IntChan[2*integ]  = (datai1 - datai2);
      /*IntChan[2*integ]  = datai1;*/
      IntChan[2*integ+1] = 0.0;        /* zero imaginary part */
      integ++;
    } /* end of data wanted section */

    /* read next row */
    read_imaging_data_table(fptr, tabledef, row, irow, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading IMAGING_DATA_TABLE\n", *ierr);
      return;
    }
    input_last = irow; /* save last row read */
    
  } /* end loop over rows */
  
  /* done? */
  if (input_last>=input_nrows) *ierr = -1;
  
  *ntime = integ;

  /* split into two? */
#ifdef DOSPLIT
  /* reverse second half */
  if (row->step<=1) {
    for (i=0; i<integ/2; i++) {
      j = integ - i;
      temp = IntChan[j*2];
      IntChan[j*2] = IntChan[i*2];
      IntChan[i*2] = temp;
      temp = IntChan[j*2+1];
      IntChan[j*2+1] = IntChan[i*2+1];
      IntChan[i*2+1] = temp;
    }
  } /* end reverse second half */
#endif
  
} /* end get_FLUOR_scan */

void FLUOR_fringe_fit_data (float *IntChan, int nt, int nscan, 
			    float time_incr,
			    float *OPD_off, float *times, 
			    imaging_detector_table_def* detector,
			    cal_table_row *cal_row)
/*----------------------------------------------------------------------- */
/*   Fit delay and rate for FLUOR data                                    */
/*   Notes:                                                               */
/*    1) setup_FOfringe_fit must be called before first call to           */
/*       FOfringe_fit_data                                                */
/*    2) shutdown_FOfringe_fit should be called after last call to        */
/*       FOfringe_fit_data                                                */
/*                                                                        */
/*   Steps:                                                               */
/*   Inputs:                                                              */
/*      IntChan   Difference array of two interferometric channels        */
/*                as complex values [scan][frame]                         */
/*      nt        Number of frames in a scan                              */
/*      nscan     Number of scans  in IntChan                             */
/*      time_incr Time step between frames in sec                         */
/*      OPD_off   OPD offset applied (seconds) [scan][frame]              */
/*      times     Times relative to reference time (sec) [scan][frame]    */
/*      detector  imaging_detector table definition structure             */
/*   Outputs:                                                             */
/*      cal_row   calibration table structure for result, ignore if NULL  */
/*----------------------------------------------------------------------- */
{
  float delay, rate, ratefact, faz, snr, centerp, xcenter, dteff, acc, s;
  float amp[256], ph[256], aamp, pph, xamp, df, dt, flip, mom[3], dOPDdt;
  float _time[256], expmom1;
  float pdelay, prate, wt1, wt2, ref_pix;
  int i, j, i1, i2, wid;
  double freq, freq0, freq_incr;

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* get frequency spacing in Fourier transformed version */
  df = 1.0 / (OPD_off[nt-1] - OPD_off[0]);
  fprintf (stdout,"Frequency spacing = %g Hz\n", df);

  /* which channel is the expected central one */
  centerp = freq0 / df;

  /* save to image if requested */
  if (FLUORimage && ((FOlast+FOnx)<=(FOmax))) {
    for (i=0; i<nt; i++) FLUORplane[FOlast++] = IntChan[i*2];
  }  /* end save to image */
  
  /* debug 
  for (i=0; i<nt; i++) {
    fprintf (stdout,"%d %7.2f\n", 
	     i, IntChan[i*2]);
	     }*/

  /* FFT to sky freq */
  FOFT_in_time(IntChan, nt);

  /* estimate rate from centroid */
  xcenter = FLUOR_find_centroid (IntChan, nt/2);
  ratefact = (centerp - xcenter) / centerp;
  rate = ratefact * ((OPD_off[nt-1] - OPD_off[0]) / (times[nt-1] - times[0]));
  rate  *= 1.0e+6 * CLIGHT; /* to microns per second */
  fprintf (stdout, "Estimated rate = %f u/s, peak = %f\n", rate, xcenter);

  /* debug */
  /*
    FLUOR_find_delay1 (IntChan, nt/2, xcenter, &delay, &aamp, &faz, &snr);    
    delay = ((delay / TWOPI) / df);
    delay = delay * 1.0e+6 * CLIGHT;  
    fprintf (stdout,"scan %d delay  = %7.2f\n", 
    i, delay);
  */

    /* split into two? */
#ifdef DOSPLIT
  /* rotate phases by 180 ramp*/
  flip = 1.0;
  for (i=0; i<nt/2; i++) {
    IntChan[i*2]   *= flip;
    IntChan[i*2+1] *= flip;
    flip *= -1.0;
  }
#endif

  /* Do fit */
  FLUOR_find_delay1 (IntChan, nt/2, xcenter, &delay, &xamp, &faz, &snr);    

  /* moment analysis */
  wid = 20;
#ifdef DOSPLIT
  wid /= 2; /* if half lag function used */
#endif
  FLUOR_moment (mom, IntChan, centerp, nt/2, wid);
  /* add correction derived from Alpha Boo */
  mom[0] += 2.5;
  /* correct acceleration for rate */
  mom[1] *= 1.0/((1.0-ratefact)*(1.0-ratefact));
  /* keep statistics */
  summom0 += mom[0];
  cntmom0++;
  if (snr>10.0) { /* ignore bad detections */
    fprintf (stdout, 
	     "time = %10.3f moments = %7.2f  %7.2f  %7.2f pixels snr %7.2f \n",
	     cal_row->time*86400.0, mom[0], mom[1], mom[2], snr);
  }
  /* convert to OPD derivatives */
  dOPDdt = ((OPD_off[nt-1] - OPD_off[0]) / (times[nt-1] - times[0]));
  rate = -dOPDdt * (mom[0]) / centerp;
  rate  *= 1.0e+6 * CLIGHT; /* to microns per second */
  dteff = 0.03; /* roughly correct for FLUOR */
  expmom1 = 9.0; /* expected FLUOR moment 1 - wild guess */
#ifdef DOSPLIT
  expmom1 /= 4.0; /* if half lag function used */
  dteff /= 2.0;
#endif
  /* smearing function width */
  if ((mom[1] - expmom1)>0.0)
      s = sqrt (mom[1] - expmom1);
  else
      s = 0.0;
  acc = (dOPDdt * s / centerp) / dteff;
  acc  *= 1.0e+6 * CLIGHT; /* to microns per second per second*/
  if ((snr>10.0) && (mom[1]>expmom1/2.0)) { /* ignore bad detections */
    fprintf (stdout, "time = %10.3f rate = %7.2f  acceleration = %7.2f  snr %7.2f \n",
	     cal_row->time*86400.0, rate, acc, snr);
  }

  /* debug */
  /* to amp and phase */  
  for (i=0; i<256; i++) {amp[i]=0.0; ph[i]=0.0;}
  for (i=0; i<nt/2; i++) {
    amp[i] = sqrt (IntChan[i*2]*IntChan[i*2] + 
		   IntChan[i*2+1]*IntChan[i*2+1]);
    ph[i] = atan2 (IntChan[i*2+1], IntChan[i*2]);
  }
 
  /* debug - apply corrections */
  faz = -faz;
  delay = delay;
  FLUOR_apply_cal(IntChan, nt/2, xcenter, delay, faz, rate, ratefact);
	    
  /* save to image if requested */ 
  /* amp in lower half and phase in upper half */
  if (FLUORimage && ((FOlast+FOnx)<=(FOmax))) {
    for (i=0; i<nt; i++) 
      if (i<nt/2) { 
	FLUORplane[FOlast++] = sqrt(IntChan[i*2]*IntChan[i*2] + 
				    IntChan[i*2+1]*IntChan[i*2+1]);
      } else { 
	FLUORplane[FOlast++] = 57.296 * atan2(IntChan[i*2+1], IntChan[i*2]); 
      }
  }  /* end save to image */
  
  /* debug */
  /* to amp and phase */  
  /* debugshit */
    /* split into two? */
#ifdef DOSPLIT
  for (i=15; i<30; i++) {
#endif
    /* split into two? */
#ifndef DOSPLIT
  for (i=30; i<60; i++) {
#endif
    aamp = sqrt (IntChan[i*2]*IntChan[i*2] + 
		 IntChan[i*2+1]*IntChan[i*2+1]);
    pph = atan2 (IntChan[i*2+1], IntChan[i*2]);
    /*    fprintf (stdout,"%d %7.2f %6.1f %7.2f %6.1f\n", 
	  i, amp[i], 57.296*ph[i], aamp, 57.296*pph);*/
  }
  
  
  /* convert delay from rad/cell to time */
  delay = ((delay / TWOPI) / df);
  
  /* convert rate to time/time */
  rate  /= 1.0e+6 * CLIGHT;
  
  /* save result if necessary */
  if (cal_row!=NULL) {
    cal_row->gain[0] = cos(faz);
    cal_row->gain[1] = -sin(faz);
    cal_row->delay   = delay;
    cal_row->rate    = rate;
    cal_row->weight  = snr;
    /* print result */
    /* convert results from seconds to microns */
    delay *= 1.0e+6 * CLIGHT;
    rate  *= 1.0e+6 * CLIGHT;
    fprintf (stdout, "time = %10.3f   delay = %7.2f rate = %8.3f amp =%7.1f  phd = %7.1f snr = %6.2f\n",
	     cal_row->time*86400.0, delay, rate, xamp, faz*57.296, snr);
  } /* end save value */
  
} /* end FLUOR_fringe_fit_data */
  
int  FLUOR_calib_data (float *IntChan, int nt, int iscan, int nscan, 
		       int chAv, int lo, int hi,
		       float OPD_range, float calDelay, float calPhase, 
		       float calRate, float ratefact, float weight,
		       visibility_row *visrow)
/*----------------------------------------------------------------------- */
/*   Apply calibration to MIDI Fourier mode (undispersed) data            */
/*   Steps:                                                               */
/*   Inputs:                                                              */
/*      IntChan   Difference array of two interferometric channels        */
/*                as complex values.                                      */
/*      nt        Number of frames in a scan                              */
/*      iscan     Which scan is this in the sequence                      */
/*      nscan     Total number of scans in the sequence                   */
/*      chAv      Number of spectral channels to average                  */
/*      chLo      First spectral channel to average                       */
/*      chHi      Highest spectral channel to average                     */
/*      OPD_range Range of modulation OPD in a scan (sec)                 */
/*      calDelay  Calibration delay (seconds)                             */
/*      calPhase  Calibration phase (radians)                             */
/*      calRate   Calibration rate*delta time (sec)                       */
/*      ratefact  Calibration rate scale compression factor               */
/*      weight    data weight for this interferogram                      */
/*   Outputs:                                                             */
/*      visrow    Visibility record, filled in on output                  */
/*   Returns                                                              */
/*      1 if there is valid data in the average and iscan=nscan, else 0   */
/*----------------------------------------------------------------------- */
{
  int i, j, ich, nchan, out=0;
  float df, delay, rate, centerp=0, xcenter;
  /* debug */
  float amp[64], ph[64], pdelay, paamp, pfaz, prate;

  /* range of channels to average */
  /* note: the number of channels needs to be coordinated with
     routine init_outdata in calibFO.c */

  /* if first scan of average then initialize */
  if (iscan<=1) {
    ich = 0; /* output spectral channel number */
    for (i=lo; i<hi; i+=chAv) {
      for (j=0; j<chAv; j++) {
	visrow->data[2*ich]   = 0.0;
	visrow->data[2*ich+1] = 0.0;
	visrow->weight[ich]   = 0.0;
      }
      ich++;
    }
  }

  /* get frequency spacing in Fourier transformed version */
  df = 1.0 / OPD_range;

  /* FFT to sky freq */
  FOFT_in_time(IntChan, nt);

  /* estimate rate from centroid */
  xcenter = FLUOR_find_centroid (IntChan, nt/2);

  /* problems: centerp xcenter, df */

  /* convert delay and rate to channel units */
  delay = df * TWOPI * calDelay;
  rate  = df * TWOPI * calRate;
  ratefact = (centerp - xcenter) / centerp;

  /* Apply calibration */
  /* phase correction for rate at reference channel*/
  FLUOR_apply_cal(IntChan, nt/2, xcenter, delay, calPhase, rate, ratefact);

  /* to amp and phase */
  /* debug*/
  for (i=0; i<nt/4; i++) {
    amp[i] = sqrt (IntChan[i*2]*IntChan[i*2] + 
		   IntChan[i*2+1]*IntChan[i*2+1]);
    ph[i] = 57.296 * atan2 (IntChan[i*2+1], IntChan[i*2]);
    fprintf (stdout,"%d %7.2f %6.1f\n", 
	     i, amp[i], ph[i]);
  } 
  /*end debug */

  /* accumulate, averaging over frequency */
  ich = 0; /* output spectral channel number */
  for (i=lo; i<hi; i+=chAv) {
    for (j=0; j<chAv; j++) {
      visrow->data[2*ich]   += weight*IntChan[(i+j)*2];
      visrow->data[2*ich+1] += weight*IntChan[(i+j)*2+1];
      visrow->weight[ich] += weight;
    }
    ich++;
  }
  nchan = ich; /* how many channels are there */

  /* if last scan of average then normalize */
  if (iscan>=nscan) {
    for (j=0; j<nchan; j++) {
      visrow->data[j*2]   /= visrow->weight[j];
      visrow->data[j*2+1] /= visrow->weight[j];
    }
    out = (visrow->weight[0]>0.0);
  }
  return out;
  } /* end FLUOR_calib_data */

void FLUOR_find_delay1 (float *spectrum, int nf, float center,
		    float *delay, float *amp, float *phase, float *snr)
/*----------------------------------------------------------------------- */
/*   Find Delay, integrated amplitude and phase from a complex spectrum   */
/*   Inputs:                                                              */
/*      spectrum  Complex spectrum                                        */
/*      nf        Number of complex elements                              */
/*      center    Expected central pixel                                  */
/*   Outputs:                                                             */
/*      delay     delay in radians per cell                               */
/*      amp       integrated amplitude                                    */
/*      phase     phase of delay flattened spectrum                       */
/*      snr       estimate of SNR                                         */
/*----------------------------------------------------------------------- */
{
#define NSEARCH 200  /* number of cells to search */
  float delta, tdelay, centerd, centerp;
  float testsnr[NSEARCH], testamp[NSEARCH], testph[NSEARCH], testdel[NSEARCH];
  float maxv, tre, tim, sum, count;
  int i, m, imax;

  /* set initial values */
  *delay = 0.0;
  *amp   = 0.0;
  *phase = 0.0;

  /* get centroid for reference */
  centerp = center;

  /* loop over test delays */
  m = NSEARCH;      /* number of attempts */
  delta = TWOPI / m; /* delay in rad per cell */
  centerd = 0.0;     /* about center */
  for (i=0; i<m; i++) {
    tdelay = centerd + (i - m/2) * delta;
    testdel[i] = tdelay;
    testsnr[i] = 
      FLUOR_int_spectrum (spectrum, nf, tdelay, centerp, 0.0, &tre, &tim);
    testph[i] = atan2(tim, tre);
    testamp[i] = sqrt(tre*tre+tim*tim);
  } /* end loop over trials */
  
  /* find peak */
  imax = -1; maxv = 0.0;
  for (i=0; i<m; i++) {
    if (maxv<testamp[i]) {
      imax = i;
      maxv = testamp[i];
    }
  }
  
  /* set final values */
  *delay = testdel[imax];
  *amp   = testamp[imax];
  *phase = testph[imax];
  *snr = testsnr[imax];
} /* end  FLUOR_find_delay1 */

float FLUOR_int_spectrum (float *spectrum, int nf, float delay, float center,
		      float phase, float *re, float *im)
/*----------------------------------------------------------------------- */
/*   Returns the snr  of a summed spectrum after a delay correction       */
/*   Inputs:                                                              */
/*      spectrum  Complex spectrum                                        */
/*      nf        Number of complex elements                              */
/*      delay     delay in radians per cell                               */
/*      center    cell number to use as the reference                     */
/*      phase     phase offset to use in radians                          */
/*   Outputs:                                                             */
/*      re        real part of summed spectrum                            */
/*      im        imaginary part of summed spectrum                       */
/*   Returns:                                                             */
/*      Amplitude of summed spectrum after applying a correction for delay*/
/*----------------------------------------------------------------------- */
{
  float out=0.0;  /* return value */
  int i, lo, hi, count, countx;
  float xre, xim, red, imd, sumre, sumim, sumxre, sumxim, xcenter;

  /* center = 0.0; debug */
  lo = 0;
  hi = nf;
  /* debug */
  if (nf>5) {
    /*    xcenter = FLUOR_find_centroid (spectrum, nf);*/
    xcenter = center;
    lo = xcenter-4.4;
    hi = xcenter+5.6;
#ifdef DOSPLIT
    lo = xcenter-1.4;
    hi = xcenter+2.6;
#endif
  }
 
  sumre  = 0.0; sumim  = 0.0; count = 0;
  sumxre = 0.0; sumxim = 0.0; countx = 0;
  for (i=0; i<nf; i++) {
    if ((i>=lo) && (i<=hi)) { /* desired */
      red = cos (phase + (i-center) * delay);
      imd = sin (phase + (i-center) * delay);
      xre = red * spectrum[i*2] - imd * spectrum[i*2+1];
      xim = red * spectrum[i*2+1] + imd * spectrum[i*2];
      sumre += xre;
      sumim += xim;
      count++;
    } else { /* out of range */
      red = cos (phase + (i-center) * delay);
      imd = sin (phase + (i-center) * delay);
      xre = red * spectrum[i*2] - imd * spectrum[i*2+1];
      xim = red * spectrum[i*2+1] + imd * spectrum[i*2];
      sumxre += xre;
      sumxim += xim;
      countx++;
    }
  } /* end loop summing spectrum */
  *re = sumre;
  *im = sumim;
  sumre /= count;
  sumim /= count;
  sumxre /= countx;
  sumxim /= countx;
  out = sqrt ((sumre*sumre + sumim*sumim) / 
	      (sumxre*sumxre + sumxim*sumxim));
  return out;
} /* end FLUOR_int_spectrum */

void FLUOR_apply_cal(float *spectrum, int nf, float center, 
		     float delay, float phase,  float rate, float ratefrac)
/*----------------------------------------------------------------------- */
/*   Applys a delay calibration to a complex spectrum                     */
/*   Inputs:                                                              */
/*      spectrum  Complex spectrum                                        */
/*      nf        Number of complex elements                              */
/*      center    cell number to use as the reference                     */
/*      delay     delay in radians per cell                               */
/*      phase     phase offset to use in radians                          */
/*      rate      derivative of delay                                     */
/*      ratefrac  pixel compression factor due to rate                    */
/*   Outputs:                                                             */
/*      spectrum  Complex spectrum with corrections applied               */
/*----------------------------------------------------------------------- */
{
  int i;
  float centerp, xre, xim, red, imd, xpix, cx[2];

  /*  center = FLUOR_find_centroid (spectrum, nf);*/
  /*fprintf (stdout, "FLUOR_apply_cal: central frequency = %f\n", center);*/
   /* center = nf / 2;central pixel */
  /*  center = 50.0; */

  for (i=0; i<nf; i++) {
    red = cos (phase + (i-center) * delay);
    imd = sin (phase + (i-center) * delay);
    xre = red * spectrum[i*2]   - imd * spectrum[i*2+1];
    xim = red * spectrum[i*2+1] + imd * spectrum[i*2];
    FLwork[i*2]   = xre;
    FLwork[i*2+1] = xim;
  } /* end loop correcting spectrum */

  /* interpolate to correct frequencies */
  for (i=0; i<nf; i++) {
    xpix = (float)i * (1.0 - ratefrac);
    cxintp (cx, FLwork, nf, 2, xpix, 0.0);
    spectrum[i*2]   = cx[0];
    spectrum[i*2+1] = cx[1];
    /* accumulate average spectrum */
    FLcount++;
    FLspectrum[i*2]   += cx[0];
    FLspectrum[i*2+1] += cx[1];
  }
} /* end FLUOR_apply_cal */

float FLUOR_find_centroid (float *array, int n)
/*----------------------------------------------------------------------- */
/*   Find centroid of amplitude in complex array                          */
/*   Inputs:                                                              */
/*      array     Measured time sequence of detector values as complex    */
/*      n         Number of elements                                      */
/*   Return:                                                              */
/*      0-rel index number of complex element                             */
/*----------------------------------------------------------------------- */
{
  int   i, imax, lo, hi;
  float amp, out=0.0;
  double sum1, sum2, maxval;

  if (n<=0) return out; /* sanity check */

  /* find peak */
  sum1 = sum2 = 0.0;
  imax = 0; maxval=-1.0;
  for (i=0; i<n; i++) {
    amp = (array[i*2]*array[i*2] + 
	   array[i*2+1]*array[i*2+1]);
    if (maxval<amp) {
      maxval = amp;
      imax = i;
    }
  }

  /* get centroid near peak */
  lo = imax - 3;
  if (lo<0) lo = 0;
  hi = imax + 3;
  if (hi>n) hi = n;
  sum1 = sum2 = 0.0;
  imax = 0; maxval=-1.0;
  for (i=lo; i<hi; i++) {
    amp = sqrt (array[i*2]*array[i*2] + 
		array[i*2+1]*array[i*2+1]);
    sum1 += amp;
    sum2 += amp * (float)i;
  }

  /* get first moment */
  if (sum1>0.0) out = sum2/sum1;
  else out = 0.0;

  return out;
} /* end  FLUOR_find_centroid */

void FLUOR_moment (float mom[3], float *array, float center, int n, int wid)
/*----------------------------------------------------------------------- */
/*   Find 1st -  3rd moments of modulus of complex array about center     */
/*   Higher moments are about the first moment.                           */
/*   Inputs:                                                              */
/*      array     Measured time sequence of detector values as complex    */
/*      center    (0-rel) pixel coordinate of center in array             */
/*      n         Number of elements                                      */
/*      wid       Number of elements about center to include              */
/*   outputs:                                                             */
/*      mom        1st -  3rd moments                                     */
/*----------------------------------------------------------------------- */
{
  int   i, imax, lo, hi;
  float amp, x, bias;
  double sum0, sum1, sum2, sum3, maxval;

  if (n<=0) return; /* sanity check */

  lo = center + 0.5 - wid/2;
  hi = lo + wid;
  if (lo<0) lo = 0;
  if (hi>n) hi = n;

  /* determine bias from off spectrum points */
  sum0 = sum1 = bias = 0.0;
  for (i=lo/2; i<lo; i++) {
    amp = sqrt (array[i*2]*array[i*2] + 
		array[i*2+1]*array[i*2+1]);
    sum0 ++;
    sum1 += amp;
  }

  for (i=hi; i<n; i++) {
    amp = sqrt (array[i*2]*array[i*2] + 
		array[i*2+1]*array[i*2+1]);
    sum0 ++;
    sum1 += amp;
  }
  if (sum0>0.0) bias = sum1 / sum0;

  /* get first moment near center */
  sum0 = sum1 = 0.0;
  for (i=lo; i<hi; i++) {
    amp = sqrt (array[i*2]*array[i*2] + 
		array[i*2+1]*array[i*2+1])
      -bias;
    x = (float)(i) - center;
    sum0 += amp;
    sum1 += amp * x;
  }

  /* get moments */
  if (sum0>0.0) {
    mom[0] = sum1 / sum0;
  } else {
    mom[0] = 0.0;
  }

  /* get higher moments near first moment */
  sum0 = sum1 = sum2 = sum3 = 0.0;
  for (i=lo; i<hi; i++) {
    amp = sqrt (array[i*2]*array[i*2] + 
		array[i*2+1]*array[i*2+1])
      -bias;
    x = (float)(i) - (center+mom[0]);
    sum0 += amp;
    sum2 += amp * x * x;
    sum3 += amp * x * x * x;
  }

  /* debug  */
  fprintf (stdout,"Moment normalization factor %7.2f\n", 
	   sum0);
  /*sum0 = 250.0;  really debug */

  /* get higher moments */
  if (sum0>0.0) {
    mom[1] = sum2 / sum0;
    mom[2] = sum3 / sum0;
  } else {
    mom[1] = 0.0;
    mom[2] = 0.0;
  }
} /* end  FLUOR_moment */

image_info* init_FLUORpix(char *pixfile, fitsfile *inputfptr,
		       int nd, int nt, 
		       float del_d, float del_t, char *obs_date, int *ierr)
/*----------------------------------------------------------------------- */
/*   If variable pixfile is set then it is used as the output FITS        */
/*   file name, otherwise no file is written.                             */
/*   This file will contain a FITS image c                                */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      pixfile     name of output Delay-rate image, NULL=>don't write    */
/*      inputfptr   input file cfitsio pointer                            */
/*      nd          number of delay pixels                                */
/*      nt          number of time pixels                                 */
/*      del_d       delay pixel increment                                 */
/*      del_t       time pixel increment                                  */
/*      obs_date    observing date                                        */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      image_info pointer, NULL indicated no file to be written.         */
/*----------------------------------------------------------------------- */
{
  char *ctype[3] = {"DELAY","SCAN","    "};
  float crota[3] = {0.0, 0.0, 0.0};
  double crval[3] = {0.0, 0.0, 0.0};
  int  i, bp, naxis[3];
  float crpix[3], crinc[3], bscale;
  long n;
  image_info *out=NULL;

  if (*ierr!=0) return NULL;     /* existing error conditions? */
  if (pixfile==NULL) return NULL;

  /* set values */
  bp = -32;  /* save as 8 bit integers */

  naxis[0] = nd; naxis[1] = nt; naxis[2] = 1; /* axis dimensions */
  crpix[0] = nd / 2 + 1.0; /* reference pixels */
  crpix[1] = nt / 2 + 1.0;
  crpix[2] = 0.0;
  crinc[0] = del_d;
  crinc[1] = del_t;
  crinc[2] = del_t;

  fprintf (stdout, "Writing delay-scan images to %s\n",pixfile);

  /* make work arrays for image plane */
  n = nd * nt;
  FLUORplane = (float*)malloc(n*sizeof(float));
  if (FLUORplane==NULL) return NULL; /* check allocation */

  FOnx = naxis[0];
  FOmax = n;
  FOlast = 0;

  /* do it */
  out = make_image (pixfile, bp, naxis, ctype, crval, crpix, crinc, crota, 
		     obs_date, ierr);
  if (out==NULL) return out; /* failed */

  /* copy history */
  copy_history (inputfptr, out->fptr, ierr);

  /* add parameters in history FUNCTION DOES NOT EXIST */
  /* history(out->fptr, ierr); */

  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d creating output delay-rate image \n", *ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  return out; /* done */
} /* end init_FLUORpix */

/* */
